long WINAPI _export WndProc(HWND hWnd, UINT message,
                            UINT wParam, LONG lParam)
/*
  USE: SHOWWPG's main window procedure.
  IN:  Standard WndProc() parameters.
   WM_CREATE  : Set normal caption
   WM_ACTIVATE: Set minimized caption
   WM_SIZE    : Update global ClientRect
   WM_COMMAND : IDM_RETRIEVE triggers GetOpenFileName()
                IDM_COPY copies to clipboard
                IDM_ABOUT shows About DlgBox
   WM_PAINT   : Traverse and display WPG file
   WM_MOUSEMOVE  : Pass along to Modify()
   WM_LBUTTONDOWN:  to size or translate
   WM_LBUTTONUP  :        "
   WM_LBUTTONDBLCLK: Pass along to Modify() to zoom in
   WM_RBUTTONDBLCLK: Pass along to Modify() to zoom out
   WM_DESTROY    : Finalize.

   hBuf is a key global variable that is non-NULL if
   and only if it is a valid memory handle to a WPG 
   file. Check hBuf to see if you have a file to show.
*/
{
  HANDLE hNewMem;   /* temp handle to file mem */
  PAINTSTRUCT ps;   /* needed in WM_PAINT */
  HDC    hDC;       /* needed in WM_PAINT */
  LPSTR  lpls;      /* ptr to last '\' in string */
  static FARPROC lpDlg;   /* for About box function */

  switch (message)
    {
    case WM_CREATE:
      /* Write default caption for main window. */
      wsprintf(Caption, "%s - (Untitled)", AppName);
      SetWindowText(hWnd, Caption);
      /* Default iconic name is just app name. */
      lstrcpy(IconicName, AppName);
      break;  /* WM_CREATE */

    case WM_ACTIVATE:
      /* When minimized, use AppName only for caption,
       * else full caption.
       */
      SetWindowText(hWnd, IsIconic(hWnd) ? IconicName :
                    Caption);
      break;  /* WM_ACTIVATE */

    case WM_SIZE:
      /* Update global ClientRect. */
      GetClientRect(hWnd, &ClientRect);
      break;  /* WM_SIZE */

    case WM_COMMAND:
      switch (wParam)
        {
        case IDM_RETRIEVE:
          /* Get file name, read into memory. */
          if ((hNewMem = LoadWPGFile(hWnd)) == NULL)
            break;
          /* If hBuf valid, free it and reset. */
          if (hBuf) GlobalFree(hBuf);
          hBuf = hNewMem;
          break;  /* IDM_RETRIEVE */

        case IDM_COPY:
          /* hBuf non-NULL means file is in memory. */
          if (hBuf)  CopyWMFToClipBoard(hWnd);   
          break;  /* IDM_COPY */

        case IDM_ABOUT:
          lpDlg=MakeProcInstance((FARPROC)AboutDlgProc,
                hInst);
          DialogBox(hInst, "AboutBox", hWnd, lpDlg);
          FreeProcInstance(lpDlg);
          break;  /* IDM_ABOUT */
        }
      break;  /* WM_COMMAND */

    case WM_PAINT:
      hDC = BeginPaint(hWnd, &ps);
      if (hBuf)  PaintWindow(hDC);
      EndPaint(hWnd, &ps);
      break;  /* WM_PAINT */

    case WM_MOUSEMOVE:
    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONDBLCLK:
      if (hBuf)
        Modify(hWnd, message, lParam);
      else
        /* SetCursor() every MOUSEMOVE if no file. */
        SetCursor(HCursor[NO_HIT]);
      break;  /* WM_MOUSEMOVE... */

    case WM_DESTROY:
      /* Free memory handle if valid. */
      if (hBuf)  GlobalFree(hBuf);
      if ((lpls = _fstrrchr(FileName, '\\')) != NULL)
        {
        /* Excise file name so only path left. */
        *lpls = 0;
        /* Write current path to SHOWWPG.INI. */
        WritePrivateProfileString("Defaults", "Path",
                             FileName, "SHOWWPG.INI");
        }
      PostQuitMessage(0);
      break;

    default:  /* Passes it on if unproccessed */
      return DefWindowProc(hWnd,message,wParam,lParam);
    }  /* switch */

  return (NULL);
}

void LOCAL Modify(HWND hWnd, UINT message, LONG lParam)
/*
  USE:  Sizing and translating tool.
  IN:   WndProc() parameters passed along.
  NOTE: State tool with route WAITING-->START_MODIFYING
   -->MODIFYING-->WAITING. START_MODIFYING is the state
   for the first WM_MOUSEMOVE after tool has started.
   There are 9 modifications: translation + 8 types of
   resizing. Which modification is triggered depends on
   which hot spot the cursor was in at WM_LBUTTONDOWN.
   Each modification has its own section in
   RestrictCursor(), DrawDraggingRect(), UpdateScale(),
   functions called respectively at WM_LBUTTONDOWN,
   WM_MOUSEMOVE, and WM_LBUTTONUP (RestrictCusor()
   initializes, DrawDraggingRect() draws interactively,
   UpdateScale() finalizes).
*/
{
  HDC hDC;           /* DC to draw on */
  static POINT pt;   /* button-down point */
  static HPEN hPen;  /* dashed pen */
  static HRGN hRgn;  /* clip region */
  static int hs;     /* hot spot index */
  static int state;  /* system state */

  switch (message)
    {
    case WM_LBUTTONDOWN:
      if (state == WAITING)
        if ((hs = PtinHotSpot(&TargetRect,
                         MAKEPOINT(lParam))) != NO_HIT)
          {
          state = START_MODIFYING;
          hPen = CreatePen(PS_DOT, 1, 0L);
          hRgn = CreateRectRgn(0, 0, ClientRect.right,
                       ClientRect.bottom-(ILINE_HT+2));
          pt = MAKEPOINT(lParam);
          RestrictCursor(hs, hWnd, &pt);
          }
      break;  /* WM_LBUTTONDOWN */

    case WM_MOUSEMOVE:
      switch (state)
        {
        case WAITING:
          /* Set Cursor for hot spot or Arrow. */
          SetCursor(HCursor[PtinHotSpot(&TargetRect,
                                  MAKEPOINT(lParam))]);
          break;
        case START_MODIFYING:
          state = MODIFYING;
          hDC = GetDC(hWnd);
          SelectObject(hDC, hRgn);
          /* Erase Framed rect (XOR mode). */
          SetROP2(hDC, R2_NOTXORPEN);
          FrameOurRect(hDC, &TargetRect);
          /* Draw rect with SOLID pen first time. */
          DrawDraggingRect(hs, hDC, pt);
          pt = MAKEPOINT(lParam);
          /* Switch to DASHED pen for next draw rect.*/
          SelectObject(hDC, hPen);
          DrawDraggingRect(hs, hDC, pt);
          ReleaseDC(hWnd, hDC);
          break;
        case MODIFYING:
          hDC = GetDC(hWnd);
          /* Draw in XOR mode with DASHED pen. */
          SetROP2(hDC, R2_NOTXORPEN);
          SelectObject(hDC, hPen);
          SelectObject(hDC, hRgn);
          /* Erase last rect. */
          DrawDraggingRect(hs, hDC, pt);
          pt = MAKEPOINT(lParam);
          /* Draw new rect. */
          DrawDraggingRect(hs, hDC, pt);
          ReleaseDC(hWnd, hDC);
          break;
        }  /*  switch (state) */
      break;  /* WM_MOUSEMOVE */

    case WM_LBUTTONUP:
      if (state==MODIFYING || state==START_MODIFYING)
        {
        DeleteObject(hPen);
        DeleteObject(hRgn);
        UpdateScale(hs, pt);
        /* Set Cursor for hot spot or Arrow. */
        SetCursor(HCursor[PtinHotSpot(&TargetRect,
                                  MAKEPOINT(lParam))]);
        /* RePaint only if modification started. */
        if (state == MODIFYING)
          InvalidateRect(hWnd, NULL, TRUE);
        state = WAITING;
        }
      break;  /* WM_LBUTTONUP */

    case WM_LBUTTONDBLCLK:
      if (state == WAITING &&
          PtInRect(&TargetRect, MAKEPOINT(lParam)))
        {
        UpdateScale(ZOOM_IN, pt);
        InvalidateRect(hWnd, NULL, TRUE);
        }
      break;  /* WM_LBUTTONDBLCLK */

    case WM_RBUTTONDBLCLK:
      if (state == WAITING &&
          PtInRect(&TargetRect, MAKEPOINT(lParam)))
        {
        UpdateScale(ZOOM_OUT, pt);
        InvalidateRect(hWnd, NULL, TRUE);
        }
      break;  /* WM_RBUTTONDBLCLK */
    }  /* switch (message) */
}

BYTE LOCAL DrawWPGRecord(HDC hDC, LPBYTE FAR *p,
                         SCALE sc)
/*
  USE:  Decipher and draw WPG record.
  IN:   hDC  = handle to DC in which to draw
        *p = ptr to start of WPG record
        sc = SCALE struct to use in drawing record
  OUT:  *p = ptr to start of next record
  RET:  Return record type.
  NOTE: p is a ptr to a ptr to the start of the WPG
   record. The ptr scans thru the record and ends by
   pointing to the next record. The actual ptr, main-
   tained by the caller, is updated as well (this is
   why we need a ptr to a ptr). Note that:

                  p = ptr to ptr
                 *p = ptr
                **p = data
*/
{
  BYTE  recType;     /* WPG record type */
  DWORD recLen;      /* WPG record length, as DWORD */

  /* Get RecordType, advance ptr. */
  recType  = *(*p)++;

  recLen = GetRecordLength(p);

  /* Process record. Each type of record being handled
   * appears as a case constant; records not listed are
   * ignored and jumped over.
   */
  switch(recType)
    {
    case WPG_FILLATTR:
      SetBrush(hDC, *p, SET_MID);
      break;
    case WPG_LINEATTR:
      SetPen(hDC, *p, SET_MID);
      break;
    case WPG_LINE:
      DrawLine(hDC, *p, sc);
      break;
    case WPG_POLYLINE:
    case WPG_POLYGON:
      DrawPoly(hDC, *p, sc, recType);
      break;
    case WPG_ELLIPSE:
      DrawEllipse(hDC, *p, sc);
      break;
    }  /* switch */

  /* Advance ptr to next record for caller. */
  *p += recLen;

  return(recType);
}

void LOCAL InitScaleParms(HANDLE hMem, LPSCALE lps,
                          LPRECT lpr)
/*
  USE:  Initialize target rect and scale parameters.
  IN:   hMem = handle to memory containing WPG data.
  OUT:  lps = ptr to scale parameters structure
        lpr = ptr to target rectangle : region on the
              screen into which image is mapped.
  NOTE: Uses global ClientRect, sets global WPGsp.
*/
{
  LPBYTE lpWPG;    /* ptr to WPG data */
  double scaleX;   /* X scale factor */
  double scaleY;   /* Y scale factor */
  int top;   /* top of target rect if loose fit - y */
  int left;  /* left of target rect if loose fit - x */

  /* Generate ptr to WPG data. */
  lpWPG = (LPBYTE) GlobalLock(hMem);

  /* If find WPG_START1 as first record, use for WPG
   * space; else use defaults set at initialization.
   */
  if (*(lpWPG + WPG_OFFS_FIRSTREC) == WPG_START1)
    WPGsp = *((LPSIZE)(lpWPG + WPG_OFFS_WPGSIZE));

  /* Unlock memory handle. */
  GlobalUnlock(hMem);

  /* Calc scale factors in both dimensions. */
  scaleX = (double) ClientRect.right  / WPGsp.cx;
  scaleY = (double) ClientRect.bottom / WPGsp.cy;

  /* Fit target rect into client rectangle with same
   * aspect ratio as WPG space, allowing for margins.
   */
  if (scaleX < scaleY)
    {
    /* TargetRect fits snugly left and right, with
     * extra space top and bottom.
     */
    scaleX *= (1.0 - 2.0 * INIT_MARGIN);
    /* Calc TargetRect with margin based on INIT_MARGIN
     * on left/right, margin = top units on top/bottom
     */
    top = (ClientRect.bottom - scaleX * WPGsp.cy) / 2;
    lpr->left = ClientRect.right * INIT_MARGIN;
    lpr->top = top;
    lpr->right = ClientRect.right * (1.0-INIT_MARGIN);
    lpr->bottom = ClientRect.bottom - top;
    /* Calc scale parms based on this TargetRect.*/
    CalcScaleParms(WPGsp, *lpr, lps);
    }
  else
    {
    /* Target rect fits snugly top and bottom, with
     * extra space left and right.
     */
    scaleY *= (1.0 - 2.0 * INIT_MARGIN);
    /* Calc TargetRect with margin based on INIT_MARGIN
     * on top/bottom, margin = left units on left/right
     */
    left = (ClientRect.right - scaleY * WPGsp.cx) / 2;
    lpr->top = ClientRect.bottom * INIT_MARGIN;
    lpr->left = left;
    lpr->bottom =ClientRect.bottom * (1.0-INIT_MARGIN);
    lpr->right = ClientRect.right - left;
    /* Calc scale parms based on this TargetRect. */
    CalcScaleParms(WPGsp, *lpr, lps);
    }
}

void LOCAL CalcScaleParms(SIZE wpgsp, RECT tr,
                          LPSCALE lps)
/*
   USE:  Calculate scale parameters.
   IN:   wpgsp = WPG space size, x = width, y = height
         tr    = target rect on screen where image goes
   OUT:  lps = ptr to struct with scale parameters
*/
{
  /* Scale equation solutions in x. */
  lps->scaleX = (double) (tr.right - tr.left)/wpgsp.cx;
  lps->offsX = tr.left;

  /* Scale equation solutions in y. */
  lps->scaleY = (double) (tr.top - tr.bottom)/wpgsp.cy;
  lps->offsY = tr.bottom;
}
